from collections import defaultdict
import pytest

from remodel.errors import AlreadyRegisteredError
from remodel.models import Model
from remodel.registry import ModelRegistry, IndexRegistry

from . import BaseTestCase


class ModelRegistryTests(BaseTestCase):
    def setUp(self):
        super(ModelRegistryTests, self).setUp()
        self.mr = ModelRegistry()

        class Artist(Model):
            pass
        self.Artist = Artist

    def test_register(self):
        self.mr.register('Artist', self.Artist)
        assert 'Artist' in self.mr._data
        assert self.mr._data['Artist'] is self.Artist

    def test_already_registered(self):
        self.mr.register('Artist', self.Artist)
        with pytest.raises(AlreadyRegisteredError):
            self.mr.register('Artist', self.Artist)

    def test_register_invalid(self):
        with pytest.raises(ValueError):
            self.mr.register('Artist', object)

    def test_unregister(self):
        self.mr.register('Artist', self.Artist)
        self.mr.unregister('Artist')
        assert 'Artist' not in self.mr._data

    def test_unregister_invalid(self):
        with pytest.raises(KeyError):
            self.mr.unregister('Artist')

    def test_unregister_unregistered(self):
        self.mr.register('Artist', self.Artist)
        self.mr.unregister('Artist')
        with pytest.raises(KeyError):
            self.mr.unregister('Artist')

    def test_get(self):
        self.mr.register('Artist', self.Artist)
        assert self.mr.get('Artist') is self.Artist

    def test_get_invalid(self):
        with pytest.raises(KeyError):
            self.mr.get('Artist')

    def test_all(self):
        self.mr.register('Artist', self.Artist)
        assert self.mr.all() == {'Artist': self.Artist}

    def test_clear(self):
        self.mr.register('Artist', self.Artist)
        self.mr.clear()
        assert self.mr._data == {}


class IndexRegistryTests(BaseTestCase):
    def setUp(self):
        super(IndexRegistryTests, self).setUp()
        self.ir = IndexRegistry()

    def test_register(self):
        self.ir.register('Artist', 'person_id')
        assert 'Artist' in self.ir._data
        assert self.ir._data['Artist'] == set(['person_id'])

    def test_register_same(self):
        self.ir.register('Artist', 'person_id')
        self.ir.register('Artist', 'person_id')
        assert 'Artist' in self.ir._data
        assert self.ir._data['Artist'] == set(['person_id'])

    def test_unregister(self):
        self.ir.register('Artist', 'person_id')
        self.ir.unregister('Artist', 'person_id')
        assert 'Artist' in self.ir._data
        assert self.ir._data['Artist'] == set()

    def test_unregister_same(self):
        self.ir.register('Artist', 'person_id')
        self.ir.unregister('Artist', 'person_id')
        self.ir.unregister('Artist', 'person_id')
        assert 'Artist' in self.ir._data
        assert self.ir._data['Artist'] == set()

    def test_get_for_model(self):
        self.ir.register('Artist', 'person_id')
        assert self.ir.get_for_model('Artist') == set(['person_id'])

    def test_get_for_inexistent_model(self):
        assert self.ir.get_for_model('Artist') == set()

    def test_get_for_unregistered_model(self):
        self.ir.register('Artist', 'person_id')
        self.ir.unregister('Artist', 'person_id')
        assert self.ir.get_for_model('Artist') == set()

    def test_get_all(self):
        self.ir.register('Artist', 'person_id')
        assert self.ir.all() == defaultdict(set, Artist=set(['person_id']))

    def test_clear(self):
        self.ir.register('Artist', 'person_id')
        self.ir.clear()
        assert self.ir._data == defaultdict(set)



import rethinkdb as r

from remodel.helpers import create_tables, drop_tables, create_indexes
from remodel.models import Model

from . import BaseTestCase, DbBaseTestCase


class CreateTablesTests(DbBaseTestCase):
    def assert_table_created(self, table):
        assert table in r.table_list().run()

    def test_new_table(self):
        class Artist(Model):
            pass

        create_tables()
        self.assert_table_created('artists')

    def test_existing_table(self):
        class Artist(Model):
            pass

        create_tables()
        create_tables()
        self.assert_table_created('artists')


class DropTablesTests(DbBaseTestCase):
    def setUp(self):
        super(DropTablesTests, self).setUp()

        class Artist(Model):
            pass

        create_tables()

    def assert_table_dropped(self, table):
        assert table not in r.table_list().run()

    def test_existing_table(self):
        drop_tables()
        self.assert_table_dropped('artists')

    def test_dropped_table(self):
        drop_tables()
        drop_tables()
        self.assert_table_dropped('artists')


class CreateIndexesTests(DbBaseTestCase):
    def assert_indexes_created(self, table, indexes):
        assert set(indexes) == set(r.table(table).index_list().run())

    def test_new_index(self):
        class Order(Model):
            belongs_to = ('Customer',)

        create_tables()
        create_indexes()
        self.assert_indexes_created('orders', ['customer_id'])

    def test_existing_index(self):
        class Order(Model):
            belongs_to = ('Customer',)

        create_tables()
        create_indexes()
        create_indexes()
        self.assert_indexes_created('orders', ['customer_id'])



from remodel.decorators import callback

from . import BaseTestCase


class CallbackTests(BaseTestCase):
    def test_decorated_function_attribute(self):
        @callback('attribute')
        def func():
            pass

        assert hasattr(func, 'attribute')
        assert func.attribute is True



import pytest

from remodel.connection import get_conn
from remodel.helpers import create_tables, create_indexes
from remodel.models import Model
from remodel.related import (HasOneDescriptor, BelongsToDescriptor,
                             HasManyDescriptor, HasAndBelongsToManyDescriptor)

from . import BaseTestCase, DbBaseTestCase

# TODO: Ensure models are saved/retrieved correctly by performing direct queries


class HasOneDescriptorTests(DbBaseTestCase):
    """
    Tests whether HasOneDescriptor correctly responds to get/set/delete actions
    in various cases
    """

    def setUp(self):
        super(HasOneDescriptorTests, self).setUp()

        class Artist(Model):
            has_one = ('Bio',)
        self.Artist = Artist

        class Bio(Model):
            pass
        self.Bio = Bio

        create_tables()
        create_indexes()

    def test_get_for_unsaved_object(self):
        a = self.Artist()
        assert a['bio'] is None
        assert a.fields._bio_cache is None

    def test_get_for_saved_object(self):
        a = self.Artist()
        a.save()
        with pytest.raises(AttributeError):
            # Cache not set by this point
            a.fields._bio_cache
        assert a['bio'] is None
        assert a.fields._bio_cache is None

    def test_get_cached(self):
        a = self.Artist()
        a.save()
        a['bio']
        # Cache set by this point
        assert a.fields._bio_cache is None
        # TODO: Test the cache was actually hit
        assert a['bio'] is None

    def test_get_cache_changed(self):
        a = self.Artist()
        a.save()
        a['bio']
        # Cache set by this point
        b = self.Bio()
        a['bio'] = b
        # Cache changed by this point
        assert a.fields._bio_cache is b
        # TODO: Test the cache was actually hit
        assert a['bio'] is b

    def test_set_invalid_instance(self):
        a = self.Artist()
        with pytest.raises(ValueError):
            a['bio'] = self.Artist()

    def test_set_none_uncached(self):
        a = self.Artist()
        a['bio'] = None
        assert a.fields._bio_cache is None

    def test_set_none_cached_with_none(self):
        a = self.Artist()
        a['bio'] = None
        # Cache set by this point
        a['bio'] = None
        assert a.fields._bio_cache is None

    def test_set_none_cached(self):
        a = self.Artist()
        a.save()
        b = self.Bio()
        a['bio'] = b
        # Cache set by this point
        a['bio'] = None
        assert 'artist_id' not in b.fields.__dict__
        assert a.fields._bio_cache is None

    def test_set_for_unsaved_object(self):
        a = self.Artist()
        b = self.Bio()
        with pytest.raises(ValueError):
            a['bio'] = b

    def test_set_valid(self):
        a = self.Artist()
        a.save()
        b = self.Bio()
        a['bio'] = b
        assert b.fields.__dict__['artist_id'] == a['id']
        assert a.fields._bio_cache is b


class BelongsToDescriptorTests(DbBaseTestCase):
    """
    Tests whether BelongsToDescriptor correctly responds to get/set/delete actions
    in various cases
    """

    def setUp(self):
        super(BelongsToDescriptorTests, self).setUp()

        class Artist(Model):
            belongs_to = ('Person',)
        self.Artist = Artist

        class Person(Model):
            pass
        self.Person = Person

        create_tables()
        create_indexes()

    def test_get_for_saved_object(self):
        a = self.Artist()
        assert a['person'] is None
        assert a.fields._person_cache is None

    def test_get_for_unsaved_object(self):
        a = self.Artist()
        a.save()
        with pytest.raises(AttributeError):
            # Cache not set by this point
            a.fields._person_cache
        assert a['person'] is None
        assert a.fields._person_cache is None

    def test_get_cached(self):
        a = self.Artist()
        a.save()
        a['person']
        # Cache set by this point
        assert a.fields._person_cache is None
        # TODO: Test the cache was actually hit
        assert a['person'] is None

    def test_get_cache_changed(self):
        a = self.Artist()
        a.save()
        a['person']
        # Cache set by this point
        p = self.Person()
        p.save()
        a['person'] = p
        # Cache changed by this point
        assert a.fields._person_cache is p
        # TODO: Test the cache was actually hit
        assert a['person'] is p

    def test_set_invalid_instance(self):
        a = self.Artist()
        with pytest.raises(ValueError):
            a['person'] = self.Artist()

    def test_set_none_with_inexistent_value(self):
        a = self.Artist()
        a['person'] = None
        assert a.fields._person_cache is None

    def test_set_none_with_existent_value(self):
        a = self.Artist()
        p = self.Person()
        p.save()
        a['person'] = p
        # Cache set by this point
        a['person'] = None
        assert 'person_id' not in a.fields.__dict__
        assert a.fields._person_cache is None

    def test_set_unsaved_object(self):
        a = self.Artist()
        p = self.Person()
        with pytest.raises(ValueError):
            a['person'] = p

    def test_set_valid(self):
        a = self.Artist()
        p = self.Person()
        p.save()
        a['person'] = p
        assert a.fields.__dict__['person_id'] == p['id']
        assert a.fields._person_cache is p


class HasManyDescriptorTests(DbBaseTestCase):
    """
    Tests whether HasManyDescriptor correctly responds to get/set/delete actions
    in various cases
    """

    def setUp(self):
        super(HasManyDescriptorTests, self).setUp()

        class Artist(Model):
            has_many = ('Song',)
        self.Artist = Artist

        class Song(Model):
            pass
        self.Song = Song

        create_tables()
        create_indexes()

    def test_get_uncached(self):
        a = self.Artist()
        a.save()
        rel_object_handler = a['songs']
        rel_object_handler_cls = a._field_handler_cls.songs.related_object_handler_cls
        assert isinstance(rel_object_handler, rel_object_handler_cls)
        assert a.fields._songs_cache is rel_object_handler

    def test_get_cached(self):
        a = self.Artist()
        a.save()
        rel_object_handler = a['songs']
        rel_object_handler_cls = a._field_handler_cls.songs.related_object_handler_cls
        # Cache set by this point
        assert isinstance(rel_object_handler, rel_object_handler_cls)
        assert a.fields._songs_cache is rel_object_handler
        # TODO: Test the cache was actually hit
        assert a['songs'] is rel_object_handler

    def test_set(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        a['songs'] = [s1, s2]
        rel_object_handler_all = a['songs'].all()
        assert len(rel_object_handler_all) == 2
        assert rel_object_handler_all[0]['id'] in [s1['id'], s2['id']]
        assert rel_object_handler_all[1]['id'] in [s1['id'], s2['id']]

    def test_set_with_existent_objects(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        s3 = self.Song()
        a['songs'] = [s1, s2]
        # Songs set by this point
        a['songs'] = [s3]
        rel_object_handler_all = a['songs'].all()
        assert len(rel_object_handler_all) == 1
        assert rel_object_handler_all[0]['id'] == s3['id']

    def test_set_with_invalid_objects(self):
        a = self.Artist()
        a.save()
        s = self.Song()
        with pytest.raises(TypeError):
            a['songs'] = s

    def test_delete(self):
        a = self.Artist()
        a.save()
        del a['songs']
        rel_object_handler_all = a['songs'].all()
        assert len(rel_object_handler_all) == 0

    def test_delete_with_existent_objects(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        a['songs'] = [s1, s2]
        # Songs set by this point
        del a['songs']
        rel_object_handler_all = a['songs'].all()
        assert len(rel_object_handler_all) == 0


class RelatedSetTests(DbBaseTestCase):
    """
    Tests whether RelatedSet correctly responds to its defined interface
    in various cases
    """

    def setUp(self):
        super(RelatedSetTests, self).setUp()

        class Artist(Model):
            has_many = ('Song',)
        self.Artist = Artist

        class Song(Model):
            pass
        self.Song = Song

        create_tables()
        create_indexes()

    def test_all_nothing_set(self):
        a = self.Artist()
        a.save()
        assert len(a['songs'].all()) == 0

    def test_all_objects_set(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        a['songs'] = [s1, s2]
        objs = a['songs'].all()
        assert len(objs) == 2
        assert objs[0]['id'] in [s1['id'], s2['id']]
        assert objs[1]['id'] in [s1['id'], s2['id']]

    def test_all_objects_deleted(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        a['songs'] = [s1, s2]
        a['songs'].remove(s1)
        objs = a['songs'].all()
        assert len(objs) == 1
        assert objs[0]['id'] == s2['id']

    def test_filter_by_ids_nothing_set(self):
        a = self.Artist()
        a.save()
        assert len(a['songs'].filter(['id'])) == 0

    def test_filter_by_kwargs_nothing_set(self):
        a = self.Artist()
        a.save()
        assert len(a['songs'].filter(id='id')) == 0

    def test_filter_by_ids_objects_set_valid_filter(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        a['songs'] = [s1, s2]
        objs = a['songs'].filter([s1['id']])
        assert len(objs) == 1
        assert objs[0]['id'] == s1['id']

    def test_filter_by_kwargs_objects_set_valid_filter(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        a['songs'] = [s1, s2]
        objs = a['songs'].filter(id=s1['id'])
        assert len(objs) == 1
        assert objs[0]['id'] == s1['id']

    def test_filter_by_ids_objects_deleted_valid_filter(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        a['songs'] = [s1, s2]
        a['songs'].remove(s1)
        objs = a['songs'].filter([s1['id']])
        assert len(objs) == 0

    def test_filter_by_kwargs_objects_deleted_valid_filter(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        a['songs'] = [s1, s2]
        a['songs'].remove(s1)
        objs = a['songs'].filter(id=s1['id'])
        assert len(objs) == 0

    def test_filter_by_ids_objects_set_invalid_filter(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        a['songs'] = [s1, s2]
        objs = a['songs'].filter(['id'])
        assert len(objs) == 0

    def test_filter_by_kwargs_objects_set_invalid_filter(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        a['songs'] = [s1, s2]
        objs = a['songs'].filter(id='id')
        assert len(objs) == 0

    def test_get_by_id(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        a['songs'] = [s1, s2]
        obj = a['songs'].get(s1['id'])
        assert obj.fields.as_dict() == s1.fields.as_dict()

    def test_get_by_kwargs(self):
        a = self.Artist()
        a.save()
        s1 = self.Song(name='My Song')
        s2 = self.Song()
        a['songs'] = [s1, s2]
        obj = a['songs'].get(name='My Song')
        assert obj.fields.as_dict() == s1.fields.as_dict()

    def test_get_by_id_inexistent(self):
        a = self.Artist()
        a.save()
        assert a['songs'].get('id') is None

    def test_get_by_kwargs_inexistent(self):
        a = self.Artist()
        a.save()
        assert a['songs'].get(name='inexistent') is None

    def test_create_object_is_saved(self):
        a = self.Artist()
        a.save()
        a['songs'].create()
        assert len(a['songs'].all()) == 1

    def test_create_object_is_returned(self):
        a = self.Artist()
        a.save()
        assert isinstance(a['songs'].create(), self.Song)

    def test_get_or_create_existent(self):
        a = self.Artist()
        a.save()
        s = a['songs'].create()
        assert a['songs'].get_or_create(s['id'])[1] is False

    def test_get_or_create_inexistent(self):
        a = self.Artist()
        a.save()
        assert a['songs'].get_or_create(name='My Song')[1] is True

    def test_count_no_objects(self):
        a = self.Artist()
        a.save()
        assert a['songs'].count() == 0

    def test_count_some_objects(self):
        a = self.Artist()
        a.save()
        a['songs'].create()
        a['songs'].create()
        assert a['songs'].count() == 2

    def test_add_with_invalid_object(self):
        a = self.Artist()
        a.save()
        with pytest.raises(TypeError):
            a['songs'].add(self.Artist())

    def test_add_with_single_object(self):
        a = self.Artist()
        a.save()
        s = self.Song()
        a['songs'].add(s)
        assert s.fields.__dict__['artist_id'] == a['id']
        assert len(a['songs'].all()) == 1

    def test_add_with_multiple_objects(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        s3 = self.Song()
        a['songs'].add(s1, s2, s3)
        assert s1.fields.__dict__['artist_id'] == a['id']
        assert s2.fields.__dict__['artist_id'] == a['id']
        assert s3.fields.__dict__['artist_id'] == a['id']
        assert len(a['songs'].all()) == 3

    def test_add_added_object(self):
        a = self.Artist()
        a.save()
        s = self.Song()
        a['songs'].add(s)
        a['songs'].add(s)
        assert s.fields.__dict__['artist_id'] == a['id']
        assert len(a['songs'].all()) == 1

    def test_remove_with_invalid_object(self):
        a = self.Artist()
        a.save()
        with pytest.raises(ValueError):
            a['songs'].remove(self.Song())

    def test_remove_with_single_object(self):
        a = self.Artist()
        a.save()
        s = self.Song()
        a['songs'] = [s]
        a['songs'].remove(s)
        assert 'artist_id' not in s.fields.__dict__
        assert len(a['songs'].all()) == 0

    def test_remove_with_multiple_objects(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        s3 = self.Song()
        a['songs'] = [s1, s2, s3]
        a['songs'].remove(s1, s2, s3)
        assert 'artist_id' not in s1.fields.__dict__
        assert 'artist_id' not in s2.fields.__dict__
        assert 'artist_id' not in s3.fields.__dict__
        assert len(a['songs'].all()) == 0

    def test_remove_removed_object(self):
        a = self.Artist()
        a.save()
        s = self.Song()
        a['songs'] = [s]
        a['songs'].remove(s)
        with pytest.raises(ValueError):
            a['songs'].remove(s)

    def test_clear_nothing_set(self):
        a = self.Artist()
        a.save()
        a['songs'].clear()
        assert len(a['songs'].all()) == 0

    def test_clear_objects_set(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        a['songs'] = [s1, s2]
        a['songs'].clear()
        assert 'artist_id' not in self.Song.get(s1['id']).fields.__dict__
        assert 'artist_id' not in self.Song.get(s2['id']).fields.__dict__
        assert len(a['songs'].all()) == 0

    def test_custom_query_correctly_handled(self):
        a = self.Artist()
        a.save()
        a['songs'].create(name='Sandstorm')
        a['songs'].create(name='Rocket')
        with get_conn() as conn:
            # order_by should be correctly handled by a['songs']
            results = list(a['songs'].order_by('name').run(conn))
        assert results[0]['name'] == 'Rocket'
        assert results[1]['name'] == 'Sandstorm'


class HasAndBelongsToManyDescriptorTests(DbBaseTestCase):
    """
    Tests whether HasAndBelongsToManyDescriptor correctly responds to get/set/delete actions
    in various cases
    """

    def setUp(self):
        super(HasAndBelongsToManyDescriptorTests, self).setUp()

        class Artist(Model):
            has_and_belongs_to_many = ('Taste',)
        self.Artist = Artist

        class Taste(Model):
            pass
        self.Taste = Taste

        create_tables()
        create_indexes()

    def test_get_uncached(self):
        a = self.Artist()
        a.save()
        rel_m2m_object_handler = a['tastes']
        rel_m2m_object_handler_cls = a._field_handler_cls.tastes.related_m2m_object_handler_cls
        assert isinstance(rel_m2m_object_handler, rel_m2m_object_handler_cls)
        assert a.fields._tastes_cache is rel_m2m_object_handler

    def test_get_cached(self):
        a = self.Artist()
        a.save()
        rel_m2m_object_handler = a['tastes']
        rel_m2m_object_handler_cls = a._field_handler_cls.tastes.related_m2m_object_handler_cls
        # Cache set by this point
        assert isinstance(rel_m2m_object_handler, rel_m2m_object_handler_cls)
        assert a.fields._tastes_cache is rel_m2m_object_handler
        # TODO: Test the cache was actually hit
        assert a['tastes'] is rel_m2m_object_handler

    def test_set(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]
        rel_m2m_object_handler_all = a['tastes'].all()
        assert len(rel_m2m_object_handler_all) == 2
        assert rel_m2m_object_handler_all[0]['id'] in [t1['id'], t2['id']]
        assert rel_m2m_object_handler_all[1]['id'] in [t1['id'], t2['id']]

    def test_set_with_existent_objects(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        t3 = self.Taste()
        t3.save()
        a['tastes'] = [t1, t2]
        # Songs set by this point
        a['tastes'] = [t3]
        rel_m2m_object_handler_all = a['tastes'].all()
        assert len(rel_m2m_object_handler_all) == 1
        assert rel_m2m_object_handler_all[0]['id'] == t3['id']

    def test_set_with_invalid_objects(self):
        a = self.Artist()
        a.save()
        t = self.Taste()
        with pytest.raises(TypeError):
            a['tastes'] = t

    def test_delete(self):
        a = self.Artist()
        a.save()
        del a['tastes']
        rel_m2m_object_handler_all = a['tastes'].all()
        assert len(rel_m2m_object_handler_all) == 0

    def test_delete_with_existent_objects(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]
        # Songs set by this point
        del a['tastes']
        rel_m2m_object_handler_all = a['tastes'].all()
        assert len(rel_m2m_object_handler_all) == 0


class RelatedM2MSetTests(DbBaseTestCase):
    """
    Tests whether RelatedM2MSet correctly responds to its defined interface
    in various cases
    """

    def setUp(self):
        super(RelatedM2MSetTests, self).setUp()

        class Artist(Model):
            has_and_belongs_to_many = ('Taste',)
        self.Artist = Artist

        class Taste(Model):
            pass
        self.Taste = Taste

        create_tables()
        create_indexes()

    def test_all_nothing_set(self):
        a = self.Artist()
        a.save()
        assert len(a['tastes'].all()) == 0

    def test_all_objects_set(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]
        objs = a['tastes'].all()
        assert len(objs) == 2
        assert objs[0]['id'] in [t1['id'], t2['id']]
        assert objs[1]['id'] in [t1['id'], t2['id']]

    def test_all_objects_deleted(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]
        a['tastes'].remove(t1)
        objs = a['tastes'].all()
        assert len(objs) == 1
        assert objs[0]['id'] == t2['id']

    def test_filter_by_ids_nothing_set(self):
        a = self.Artist()
        a.save()
        assert len(a['tastes'].filter(['id'])) == 0

    def test_filter_by_kwargs_nothing_set(self):
        a = self.Artist()
        a.save()
        assert len(a['tastes'].filter(id='id')) == 0

    def test_filter_by_ids_objects_set_valid_filter(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]
        objs = a['tastes'].filter([t1['id']])
        assert len(objs) == 1
        assert objs[0]['id'] == t1['id']

    def test_filter_by_kwargs_objects_set_valid_filter(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]
        objs = a['tastes'].filter(id=t1['id'])
        assert len(objs) == 1
        assert objs[0]['id'] == t1['id']

    def test_filter_by_ids_objects_deleted_valid_filter(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]
        a['tastes'].remove(t1)
        objs = a['tastes'].filter([t1['id']])
        assert len(objs) == 0

    def test_filter_by_kwargs_objects_deleted_valid_filter(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]
        a['tastes'].remove(t1)
        objs = a['tastes'].filter(id=t1['id'])
        assert len(objs) == 0

    def test_filter_by_ids_objects_set_invalid_filter(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]
        objs = a['tastes'].filter(['id'])
        assert len(objs) == 0

    def test_filter_by_kwargs_objects_set_invalid_filter(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]
        objs = a['tastes'].filter(id='id')
        assert len(objs) == 0

    def test_get_by_id(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]
        obj = a['tastes'].get(t1['id'])
        assert obj.fields.as_dict() == t1.fields.as_dict()

    def test_get_by_kwargs(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste(name='My Taste')
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]
        obj = a['tastes'].get(name='My Taste')
        assert obj.fields.as_dict() == t1.fields.as_dict()

    def test_get_by_id_inexistent(self):
        a = self.Artist()
        a.save()
        assert a['tastes'].get('id') is None

    def test_get_by_kwargs_inexistent(self):
        a = self.Artist()
        a.save()
        assert a['tastes'].get(name='inexistent') is None

    def test_create_object_is_saved(self):
        a = self.Artist()
        a.save()
        a['tastes'].create()
        assert len(a['tastes'].all()) == 1

    def test_create_object_is_returned(self):
        a = self.Artist()
        a.save()
        assert isinstance(a['tastes'].create(), self.Taste)

    def test_get_or_create_existent(self):
        a = self.Artist()
        a.save()
        t = a['tastes'].create()
        assert a['tastes'].get_or_create(t['id'])[1] is False

    def test_get_or_create_inexistent(self):
        a = self.Artist()
        a.save()
        assert a['tastes'].get_or_create(name='My Taste')[1] is True

    def test_count_no_objects(self):
        a = self.Artist()
        a.save()
        assert a['tastes'].count() == 0

    def test_count_some_objects(self):
        a = self.Artist()
        a.save()
        a['tastes'].create()
        a['tastes'].create()
        assert a['tastes'].count() == 2

    def test_add_with_invalid_object(self):
        a = self.Artist()
        a.save()
        with pytest.raises(TypeError):
            a['tastes'].add(self.Artist())

    def test_add_with_unsaved_object(self):
        a = self.Artist()
        a.save()
        with pytest.raises(ValueError):
            a['tastes'].add(self.Taste())

    def test_add_with_single_object(self):
        a = self.Artist()
        a.save()
        t = self.Taste()
        t.save()
        a['tastes'].add(t)
        assert len(a['tastes'].all()) == 1

    def test_add_with_multiple_objects(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        t3 = self.Taste()
        t3.save()
        a['tastes'].add(t1, t2, t3)
        assert len(a['tastes'].all()) == 3

    def test_add_added_object(self):
        a = self.Artist()
        a.save()
        t = self.Taste()
        t.save()
        a['tastes'].add(t)
        a['tastes'].add(t)
        assert len(a['tastes'].all()) == 1

    def test_remove_with_invalid_object(self):
        a = self.Artist()
        a.save()
        with pytest.raises(TypeError):
            a['tastes'].remove(self.Artist())

    def test_remove_with_single_object(self):
        a = self.Artist()
        a.save()
        t = self.Taste()
        t.save()
        a['tastes'] = [t]
        a['tastes'].remove(t)
        assert len(a['tastes'].all()) == 0

    def test_remove_with_multiple_objects(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        t3 = self.Taste()
        t3.save()
        a['tastes'] = [t1, t2, t3]
        a['tastes'].remove(t1, t2, t3)
        assert len(a['tastes'].all()) == 0

    def test_remove_removed_object(self):
        a = self.Artist()
        a.save()
        t = self.Taste()
        t.save()
        a['tastes'] = [t]
        a['tastes'].remove(t)
        a['tastes'].remove(t)
        assert len(a['tastes'].all()) == 0

    def test_clear_nothing_set(self):
        a = self.Artist()
        a.save()
        a['tastes'].clear()
        assert len(a['tastes'].all()) == 0

    def test_clear_objects_set(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]
        a['tastes'].clear()
        assert len(a['tastes'].all()) == 0

    def test_custom_query_correctly_handled(self):
        a = self.Artist()
        a.save()
        a['tastes'].create(name='House')
        a['tastes'].create(name='Classical')
        with get_conn() as conn:
            # order_by should be correctly handled by a['tastes']
            results = list(a['tastes'].order_by('name').run(conn))
        assert results[0]['name'] == 'Classical'
        assert results[1]['name'] == 'House'



from threading import Lock
from warnings import warn
from .decorators import synchronized


def create_tables():
    from .helpers import create_tables as ct
    deprecation_warning('remodel.utils.create_tables will be deprecated soon. '
                        'Please use remodel.helpers.create_tables instead.')
    ct()


def create_indexes():
    from .helpers import create_indexes as ci
    deprecation_warning('remodel.utils.create_indexes will be deprecated soon. '
                        'Please use remodel.helpers.create_indexes instead.')
    ci()


class Counter(object):
    lock = Lock()

    def __init__(self, init=0):
        self.n = init

    @synchronized(lock)
    def incr(self):
        self.n += 1

    @synchronized(lock)
    def decr(self):
        self.n -= 1

    @synchronized(lock)
    def current(self):
        return self.n


def deprecation_warning(message):
    warn(message, DeprecationWarning, stacklevel=2)



import os
import rethinkdb as r
from rethinkdb.errors import RqlDriverError
import unittest

from remodel.connection import pool, get_conn
from remodel.helpers import create_tables
from remodel.models import Model
from remodel.registry import model_registry, index_registry


def get_env_settings():
    settings = {}
    settings['host'] = os.environ.get('RETHINKDB_HOST', None)
    settings['port'] = os.environ.get('RETHINKDB_PORT', None)
    settings['auth_key'] = os.environ.get('RETHINKDB_AUTH_KEY', None)
    settings['db'] = os.environ.get('RETHINKDB_DB', None)
    return {k: v for k, v in settings.items() if v is not None}


pool.configure(max_connections=1, **get_env_settings())


class BaseTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        model_registry.clear()
        index_registry.clear()


class DbBaseTestCase(BaseTestCase):
    def setUp(self):
        super(DbBaseTestCase, self).setUp()

        # Create new database and switch to it
        try:
            with get_conn() as conn:
                r.db_create('testing').run(conn)
                conn.use('testing')
        except RqlDriverError:
            raise unittest.SkipTest('RethinkDB is unaccessible')

    def tearDown(self):
        super(DbBaseTestCase, self).tearDown()

        # Drop test database
        with get_conn() as conn:
            r.db_drop('testing').run(conn)




import rethinkdb as r
from contextlib import contextmanager
try:
    from queue import Queue, Empty
except ImportError:
    from Queue import Queue, Empty


from .utils import Counter


class Connection(object):
    def __init__(self, db='test', host='localhost', port=28015, auth_key=''):
        self.db = db
        self.host = host
        self.port = port
        self.auth_key = auth_key
        self._conn = None

    def connect(self):
        self._conn = r.connect(host=self.host, port=self.port,
                               auth_key=self.auth_key, db=self.db)

    def close(self):
        if self._conn:
            self._conn.close()
            self._conn = None

    @property
    def conn(self):
        if not self._conn:
            self.connect()
        return self._conn


class ConnectionPool(object):
    def __init__(self, max_connections=5):
        self.q = Queue()
        self.max_connections = max_connections
        self._created_connections = Counter()
        self.connection_class = Connection
        self.connection_kwargs = {}

    def configure(self, max_connections=5, **connection_kwargs):
        self.max_connections = max_connections
        self.connection_kwargs = connection_kwargs

    def get(self):
        try:
            return self.q.get_nowait()
        except Empty:
            if self._created_connections.current() < self.max_connections:
                conn = self.connection_class(**self.connection_kwargs).conn
                self._created_connections.incr()
                return conn
            raise

    def put(self, connection):
        self.q.put(connection)
        self._created_connections.decr()

    def created(self):
        return self._created_connections.current()


pool = ConnectionPool()


@contextmanager
def get_conn():
    conn = pool.get()
    try:
        yield conn
    finally:
        pool.put(conn)




import rethinkdb as r
from inflection import tableize

from .decorators import cached_property
from .object_handler import ObjectHandler
from .registry import model_registry


class RelationDescriptor(object):
    @property
    def model_cls(self):
        return model_registry.get(self.model)


class HasOneDescriptor(RelationDescriptor):
    def __init__(self, model, lkey, rkey):
        self.model = model
        self.lkey = lkey
        self.rkey = rkey
        self.related_cache = '_%s_cache' % model.lower()

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        try:
            return getattr(instance, self.related_cache)
        except AttributeError:
            instance_lkey = getattr(instance, self.lkey, None)
            if instance_lkey is None:
                rel_obj = None
            else:
                params = {self.rkey: instance_lkey}
                rel_obj = self.model_cls.get(**params)
            # Make related document available on parent (this) e.g.: user.profile
            setattr(instance, self.related_cache, rel_obj)
            return rel_obj

    def __set__(self, instance, value):
        if value is not None and not isinstance(value, self.model_cls):
            raise ValueError('%s instance expected, got %r' % (
                            self.model_cls.__name__, value))

        if value is None:
            rel_obj = getattr(instance, self.related_cache, None)
            if rel_obj is not None:
                # We are deleting the rkey attr on related field handler, not obj
                del rel_obj.fields.__dict__[self.rkey]
        else:
            instance_lkey = getattr(instance, self.lkey, None)
            if instance_lkey is None:
                raise ValueError('Cannot assign "%r": current instance isn\'t '
                                 'saved' % value)
            # Assign field this way to skip validation
            value.fields.__dict__[self.rkey] = instance_lkey
        # Make related document available on parent (this) e.g.: user.profile
        setattr(instance, self.related_cache, value)

    def __delete__(self, instance):
        self.__set__(instance, None)


class BelongsToDescriptor(RelationDescriptor):
    def __init__(self, model, lkey, rkey):
        self.model = model
        self.lkey = lkey
        self.rkey = rkey
        self.related_cache = '_%s_cache' % model.lower()

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        try:
            return getattr(instance, self.related_cache)
        except AttributeError:
            instance_lkey = instance.__dict__.get(self.lkey, None)
            if instance_lkey is None:
                rel_obj = None
            else:
                params = {self.rkey: instance_lkey}
                rel_obj = self.model_cls.get(**params)
            # Make parent document available on related (this) e.g.: profile.user
            setattr(instance, self.related_cache, rel_obj)
            return rel_obj

    def __set__(self, instance, value):
        if value is not None and not isinstance(value, self.model_cls):
            raise ValueError('%s instance expected, got %r' % (
                            self.model_cls.__name__, value))

        if value is None:
            if self.lkey in instance.__dict__:
                del instance.__dict__[self.lkey]
        else:
            value_rkey = getattr(value.fields, self.rkey, None)
            if value_rkey is None:
                raise ValueError('Cannot assign "%r": "%s" instance isn\'t '
                                 'saved' % (value, value.__class__.__name__))
            # Assign field this way to skip validation
            instance.__dict__[self.lkey] = value_rkey
        # Make parent document available on related (this) e.g.: profile.user
        setattr(instance, self.related_cache, value)

    def __delete__(self, instance):
        self.__set__(instance, None)


def create_related_object_handler_cls(model_cls, lkey, rkey):
    class RelatedObjectHandler(ObjectHandler):
        def __init__(self, parent):
            super(RelatedObjectHandler, self).__init__(model_cls)
            # Parent field handler instance
            self.parent = parent
            self.query = self.query.get_all(self._get_parent_lkey(), index=rkey)

        def create(self, **kwargs):
            obj = super(RelatedObjectHandler, self).create(**kwargs)
            self.add(obj)
            return obj

        def get_or_create(self, id_=None, **kwargs):
            obj, created = super(RelatedObjectHandler, self).get_or_create(id_, **kwargs)
            self.add(obj)
            return obj, created

        def add(self, *objs):
            for obj in objs:
                if not isinstance(obj, model_cls):
                    raise TypeError('%s instance expected, got %r' %
                                    (model_cls.__name__, obj))
                # Assign field this way to skip validation
                obj.fields.__dict__[rkey] = self._get_parent_lkey()
                obj.save()

        def remove(self, *objs):
            ref_key = self._get_parent_lkey()
            for obj in objs:
                obj_key = obj.fields.__dict__.get(rkey, None)
                if obj_key != ref_key:
                    raise ValueError('%r is not a related object' % obj)
                del obj.fields.__dict__[rkey]
                obj.save()

        def clear(self):
            for obj in self.all():
                del obj.fields.__dict__[rkey]
                obj.save()

        def _get_parent_lkey(self):
            parent_lkey = getattr(self.parent, lkey, None)
            if parent_lkey is None:
                raise ValueError('Cannot access related "%s" set: current '
                                 'instance isn\'t saved' % model_cls.__name__)
            return parent_lkey

    return RelatedObjectHandler


class HasManyDescriptor(RelationDescriptor):
    def __init__(self, model, lkey, rkey):
        self.model = model
        self.lkey = lkey
        self.rkey = rkey
        self.related_cache = '_%s_cache' % tableize(model)

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        try:
            return getattr(instance, self.related_cache)
        except AttributeError:
            rel_object_handler = self.related_object_handler_cls(instance)
            # Make related set available on parent (this) e.g.: artist.songs
            setattr(instance, self.related_cache, rel_object_handler)
            return rel_object_handler

    def __set__(self, instance, value):
        rel_object_handler = self.__get__(instance)
        rel_object_handler.clear()
        rel_object_handler.add(*value)

    def __delete__(self, instance):
        rel_object_handler = self.__get__(instance)
        rel_object_handler.clear()

    @cached_property
    def related_object_handler_cls(self):
        return create_related_object_handler_cls(self.model_cls, self.lkey, self.rkey)


def create_related_m2m_object_handler_cls(model_cls, lkey, rkey, join_model_cls, mlkey, mrkey):
    class RelatedM2MObjectHandler(ObjectHandler):
        def __init__(self, parent):
            super(RelatedM2MObjectHandler, self).__init__(model_cls)
            # Parent field handler instance
            self.parent = parent
            # Returns all docs from model_cls which are referenced in join_model_cls
            self.query = (r.table(join_model_cls._table)
                          .get_all(self._get_parent_lkey(), index=mlkey)
                          .eq_join(mrkey, r.table(model_cls._table), index=rkey)
                          .map(lambda res: res['right']))

        def create(self, **kwargs):
            obj = super(RelatedM2MObjectHandler, self).create(**kwargs)
            self.add(obj)
            return obj

        def get_or_create(self, id_=None, **kwargs):
            obj, created = super(RelatedM2MObjectHandler, self).get_or_create(id_, **kwargs)
            self.add(obj)
            return obj, created

        def add(self, *objs):
            new_keys = set()
            for obj in objs:
                if not isinstance(obj, model_cls):
                    raise TypeError('%s instance expected, got %r' %
                                    (model_cls.__name__, obj))
                obj_key = getattr(obj.fields, rkey, None)
                if obj_key is None:
                    raise ValueError('Cannot add %r: the value for field %s '
                                     'is missing (try saving the object first'
                                     ')' % (obj, rkey))
                new_keys.add(obj_key)

            existing_keys = {doc[rkey]
                            for doc in self.query.run()}
            new_keys -= existing_keys

            for obj_key in new_keys:
                params = {mlkey: self._get_parent_lkey(),
                          mrkey: obj_key}
                join_model_cls.table.insert(params).run()

        def remove(self, *objs):
            old_keys = set()
            for obj in objs:
                if not isinstance(obj, model_cls):
                    raise TypeError('%s instance expected, got %r' %
                                    (model_cls.__name__, obj))
                obj_key = getattr(obj.fields, rkey, None)
                if obj_key is not None:
                    old_keys.add(obj_key)

            existing_keys = {doc[rkey]
                            for doc in self.query.run()}
            # Remove inexisting keys from old_keys
            old_keys &= existing_keys

            if old_keys:
                (join_model_cls.table.get_all(r.args(list(old_keys)), index=mrkey)
                                     .delete()
                                     .run())

        def clear(self):
            (join_model_cls.table
             .get_all(self._get_parent_lkey(), index=mlkey)
             .delete().run())

        def _get_parent_lkey(self):
            parent_lkey = getattr(self.parent, lkey, None)
            if parent_lkey is None:
                raise ValueError('Cannot access related "%s" set: current '
                                 'instance isn\'t saved' %  model_cls.__name__)
            return parent_lkey

    return RelatedM2MObjectHandler


class HasAndBelongsToManyDescriptor(RelationDescriptor):
    def __init__(self, model, lkey, rkey, join_model, mlkey, mrkey):
        self.model = model
        self.lkey = lkey
        self.rkey = rkey
        self.join_model = join_model
        self.mlkey = mlkey
        self.mrkey = mrkey
        self.related_cache = '_%s_cache' % tableize(model)

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        try:
            return getattr(instance, self.related_cache)
        except AttributeError:
            rel_m2m_object_handler = self.related_m2m_object_handler_cls(instance)
            # Make related set available on parent (this) e.g.: user.artists
            setattr(instance, self.related_cache, rel_m2m_object_handler)
            return rel_m2m_object_handler

    def __set__(self, instance, value):
        rel_m2m_object_handler = self.__get__(instance)
        rel_m2m_object_handler.clear()
        rel_m2m_object_handler.add(*value)

    def __delete__(self, instance):
        rel_m2m_object_handler = self.__get__(instance)
        rel_m2m_object_handler.clear()

    @cached_property
    def related_m2m_object_handler_cls(self):
        return create_related_m2m_object_handler_cls(
            self.model_cls, self.lkey, self.rkey,
            self.join_model_cls, self.mlkey, self.mrkey)

    @property
    def join_model_cls(self):
        return model_registry.get(self.join_model)



import pytest
import unittest

from remodel.helpers import create_tables, create_indexes
from remodel.models import Model
from remodel.registry import index_registry
from remodel.related import (HasOneDescriptor, BelongsToDescriptor,
                             HasManyDescriptor, HasAndBelongsToManyDescriptor)

from . import BaseTestCase, DbBaseTestCase


class RelationDeclarationTests(BaseTestCase):
    """
    Tests if relations are correctly declared or not
    """

    def test_valid_relation(self):
        class Artist(Model):
            has_one = ('Bio',)

    def test_invalid_relation(self):
        with pytest.raises(ValueError):
            class Artist(Model):
                has_one = 'Bio'
        with pytest.raises(ValueError):
            class Artist(Model):
                has_one = ['Bio']
        with pytest.raises(ValueError):
            class Artist(Model):
                has_one = list
        with pytest.raises(NameError):
            class Artist(Model):
                has_one = Bio

    def test_empty_relation(self):
        class Artist(Model):
            has_one = ()

    def test_no_relation(self):
        class Artist(Model):
            pass

    def test_multiple_relation(self):
        class Artist(Model):
            has_one = ('Bio', 'Profile')

    def test_same_multiple_relation(self):
        """
        No error should be thrown; instead create a single relation
        """

        class Artist(Model):
            has_one = ('Bio', 'Bio')


class AttributesTests(BaseTestCase):
    """
    Tests whether field handler attributes are correctly set for given model
    w/ and w/o relations
    """

    def test_removed_attributes(self):
        class Artist(Model):
            has_one = ('Bio',)
            belongs_to = ('Person',)
            has_many = ('Song',)
            has_and_belongs_to_many = ('Taste',)

        fhcls = Artist._field_handler_cls
        assert not hasattr(fhcls, 'model')
        assert not hasattr(fhcls, 'has_one')
        assert not hasattr(fhcls, 'belongs_to')
        assert not hasattr(fhcls, 'has_many')
        assert not hasattr(fhcls, 'has_and_belongs_to_many')

    def test_no_relation(self):
        class Artist(Model):
            pass

        fhcls = Artist._field_handler_cls
        assert fhcls.restricted == set()
        assert fhcls.related == set()

    def test_has_one(self):
        class Artist(Model):
            has_one = ('Bio',)

        fhcls = Artist._field_handler_cls
        assert isinstance(fhcls.bio, HasOneDescriptor)
        assert fhcls.restricted == set()
        assert fhcls.related == set(['bio'])

    def test_belongs_to(self):
        class Artist(Model):
            belongs_to = ('Person',)

        fhcls = Artist._field_handler_cls
        assert isinstance(fhcls.person, BelongsToDescriptor)
        assert fhcls.restricted == set(['person_id'])
        assert fhcls.related == set(['person'])

    def test_has_many(self):
        class Artist(Model):
            has_many = ('Song',)

        fhcls = Artist._field_handler_cls
        assert isinstance(fhcls.songs, HasManyDescriptor)
        assert fhcls.restricted == set()
        assert fhcls.related == set(['songs'])

    def test_has_and_belongs_to_many(self):
        class Artist(Model):
            has_and_belongs_to_many = ('Taste',)

        fhcls = Artist._field_handler_cls
        assert isinstance(fhcls.tastes, HasAndBelongsToManyDescriptor)
        assert fhcls.restricted == set()
        assert fhcls.related == set(['tastes'])

    def test_all_relations(self):
        class Artist(Model):
            has_one = ('Bio',)
            belongs_to = ('Person',)
            has_many = ('Song',)
            has_and_belongs_to_many = ('Taste',)

        fhcls = Artist._field_handler_cls
        assert isinstance(fhcls.bio, HasOneDescriptor)
        assert isinstance(fhcls.person, BelongsToDescriptor)
        assert isinstance(fhcls.songs, HasManyDescriptor)
        assert isinstance(fhcls.tastes, HasAndBelongsToManyDescriptor)
        assert fhcls.restricted == set(['person_id'])
        assert fhcls.related == set(['bio', 'person', 'songs', 'tastes'])


class IndexesTests(BaseTestCase):
    """
    Tests whether indexes are set on the correct tables and keys
    """

    def test_has_one(self):
        class Bear(Model):
            has_one = ('FavoriteCub',)

        assert index_registry.get_for_model('Bear') == set()
        assert index_registry.get_for_model('FavoriteCub') == set(['bear_id'])

    def test_belongs_one(self):
        class Bear(Model):
            belongs_to = ('Family',)

        assert index_registry.get_for_model('Bear') == set(['family_id'])
        assert index_registry.get_for_model('Family') == set()

    def test_has_many(self):
        class Bear(Model):
            has_many = ('Cub',)

        assert index_registry.get_for_model('Bear') == set()
        assert index_registry.get_for_model('Cub') == set(['bear_id'])

    def test_has_and_belongs_to_many(self):
        class Bear(Model):
            has_and_belongs_to_many = ('Continent',)

        assert index_registry.get_for_model('Bear') == set()
        assert index_registry.get_for_model('Continent') == set()
        assert index_registry.get_for_model('_BearContinent') == set(['bear_id', 'continent_id'])

    def test_all_relations(self):
        class Bear(Model):
            has_one = ('FavoriteCub',)
            belongs_to = ('Family',)
            has_many = ('Cub',)
            has_and_belongs_to_many = ('Continent',)

        assert index_registry.get_for_model('Bear') == set(['family_id'])
        assert index_registry.get_for_model('FavoriteCub') == set(['bear_id'])
        assert index_registry.get_for_model('Family') == set()
        assert index_registry.get_for_model('Cub') == set(['bear_id'])
        assert index_registry.get_for_model('Continent') == set()
        assert index_registry.get_for_model('_BearContinent') == set(['bear_id', 'continent_id'])

class AttributeAccessTests(BaseTestCase):
    """
    Tests whether access is correctly granted to attributes
    """

    def test_normal_field(self):
        class Artist(Model):
            has_one = ('Bio',)
            belongs_to = ('Person',)

        a = Artist()
        a['name'] = 'Andrei'
        a['name']
        del a['name']

    def test_restricted_field(self):
        class Artist(Model):
            belongs_to = ('Person',)

        a = Artist()
        with pytest.raises(KeyError):
            a['person_id'] = 1
        with pytest.raises(KeyError):
            a['person_id']
        with pytest.raises(KeyError):
            del a['person_id']

    def test_relation_field(self):
        class Artist(Model):
            has_one = ('Bio',)

        class Bio(Model):
            pass

        a = Artist()
        with pytest.raises(ValueError):
            # Fails because instance must be saved before assignment
            a['bio'] = Bio()
        a['bio']
        del a['bio']


class AsDictTests(DbBaseTestCase):
    def setUp(self):
        super(AsDictTests, self).setUp()

        class Artist(Model):
            belongs_to = ('Person',)
        self.Artist = Artist

        class Person(Model):
            pass
        self.Person = Person

        create_tables()
        create_indexes()

    def test_normal_fields(self):
        a = self.Artist(name='Andrei')
        a['country'] = 'Romania'
        assert a.fields.as_dict() == {'name': 'Andrei', 'country': 'Romania'}

    def test_belongs_to(self):
        p = self.Person()
        p.save()
        a = self.Artist(person=p)
        assert a.fields.as_dict() == {'person_id': p['id']}



from inflection import tableize

from .errors import AlreadyRegisteredError
import remodel.models
from .registry import index_registry
from .related import (HasOneDescriptor, BelongsToDescriptor, HasManyDescriptor,
                     HasAndBelongsToManyDescriptor)


class FieldHandlerBase(type):
    def __new__(cls, name, bases, dct):
        if not all(isinstance(dct[rel_type], tuple) for rel_type in remodel.models.REL_TYPES):
            raise ValueError('Related models must be passed as a tuple')

        # TODO: Find a way to pass model class to its field handler class
        model = dct.pop('model')
        dct['restricted'], dct['related'] = set(), set()
        for rel in dct.pop('has_one'):
            if isinstance(rel, tuple):
                # 4-tuple relation supplied
                other, field, lkey, rkey = rel
            else:
                # Just the related model supplied
                other = rel
                field, lkey, rkey = other.lower(), 'id', '%s_id' % model.lower()
            dct[field] = HasOneDescriptor(other, lkey, rkey)
            dct['related'].add(field)
            index_registry.register(other, rkey)
        for rel in dct.pop('belongs_to'):
            if isinstance(rel, tuple):
                other, field, lkey, rkey = rel
            else:
                other = rel
                field, lkey, rkey = other.lower(), '%s_id' % other.lower(), 'id'
            dct[field] = BelongsToDescriptor(other, lkey, rkey)
            dct['related'].add(field)
            dct['restricted'].add(lkey)
            index_registry.register(model, lkey)
        for rel in dct.pop('has_many'):
            if isinstance(rel, tuple):
                other, field, lkey, rkey = rel
            else:
                other = rel
                field, lkey, rkey = tableize(other), 'id', '%s_id' % model.lower()
            dct[field] = HasManyDescriptor(other, lkey, rkey)
            dct['related'].add(field)
            index_registry.register(other, rkey)
        for rel in dct.pop('has_and_belongs_to_many'):
            if isinstance(rel, tuple):
                other, field, lkey, rkey = rel
            else:
                other = rel
                field, lkey, rkey = tableize(other), 'id', 'id'
            join_model = '_' + ''.join(sorted([model, other]))
            try:
                remodel.models.ModelBase(join_model, (remodel.models.Model,), {})
            except AlreadyRegisteredError:
                # HABTM join_model model has been registered, probably from the
                # other end of the relation
                pass
            mlkey, mrkey = '%s_id' % model.lower(), '%s_id' % other.lower()
            dct[field] = HasAndBelongsToManyDescriptor(other, lkey, rkey, join_model, mlkey, mrkey)
            dct['related'].add(field)
            index_registry.register(join_model, mlkey)
            index_registry.register(join_model, mrkey)

        return super(FieldHandlerBase, cls).__new__(cls, name, bases, dct)


class FieldHandler(object):
    def __getattribute__(self, name):
        if name in super(FieldHandler, self).__getattribute__('restricted'):
            raise AttributeError('Cannot access %s: field is restricted' % name)
        return super(FieldHandler, self).__getattribute__(name)

    def __setattr__(self, name, value):
        if name in self.restricted:
            raise AttributeError('Cannot set %s: field is restricted' % name)
        super(FieldHandler, self).__setattr__(name, value)

    def __delattr__(self, name):
        if name in self.restricted:
            raise AttributeError('Cannot delete %s: field is restricted' % name)
        super(FieldHandler, self).__delattr__(name)

    def as_dict(self):
        return {field: self.__dict__[field] for field in self.__dict__
                if not field.startswith('_')}



import pytest
import rethinkdb as r
import unittest

from remodel.connection import get_conn
from remodel.errors import OperationError
from remodel.helpers import create_tables, create_indexes
from remodel.models import Model
from remodel.object_handler import ObjectHandler, ObjectSet
from remodel.registry import model_registry
from remodel.related import (HasOneDescriptor, BelongsToDescriptor,
                             HasManyDescriptor, HasAndBelongsToManyDescriptor)

from . import BaseTestCase, DbBaseTestCase


class AllTests(DbBaseTestCase):
    def setUp(self):
        super(AllTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

        create_tables()
        create_indexes()

    def test_returns_object_set(self):
        assert isinstance(self.Artist.all(), ObjectSet)

    def test_no_objects(self):
        assert len(self.Artist.all()) == 0

    def test_some_objects(self):
        self.Artist.create()
        self.Artist.create()
        assert len(self.Artist.all()) == 2

    def test_some_objects_deleted(self):
        a = self.Artist.create()
        self.Artist.create()
        a.delete()
        assert len(self.Artist.all()) == 1


class CreateTests(DbBaseTestCase):
    def setUp(self):
        super(CreateTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

        create_tables()
        create_indexes()

    def test_object_is_saved(self):
        self.Artist.create()
        assert len(self.Artist.all()) == 1

    def test_object_is_returned(self):
        assert isinstance(self.Artist.create(), self.Artist)


class GetTests(DbBaseTestCase):
    def setUp(self):
        super(GetTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

        create_tables()
        create_indexes()

    def test_by_id(self):
        a = self.Artist()
        a.save()
        assert self.Artist.get(a['id']).fields.as_dict() == a.fields.as_dict()

    def test_by_kwargs(self):
        a = self.Artist(name='Andrei', country='Romania')
        a.save()
        assert self.Artist.get(name='Andrei', country='Romania').fields.as_dict() == a.fields.as_dict()

    def test_by_id_inexistent(self):
        assert self.Artist.get('id') is None

    def test_by_kwargs_inexistent(self):
        assert self.Artist.get(name='inexistent') is None


class GetOrCreateTests(DbBaseTestCase):
    def setUp(self):
        super(GetOrCreateTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

        create_tables()
        create_indexes()

    def test_existent(self):
        a = self.Artist.create()
        assert self.Artist.get_or_create(a['id'])[1] is False

    def test_inexistent(self):
        assert self.Artist.get_or_create(name='Andrei')[1] is True


class FilterTests(DbBaseTestCase):
    def setUp(self):
        super(FilterTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

        create_tables()
        create_indexes()

    def test_returns_object_set(self):
        assert isinstance(self.Artist.filter(), ObjectSet)

    def test_by_ids_no_objects(self):
        assert len(self.Artist.filter(['id'])) == 0

    def test_by_kwargs_no_objects(self):
        assert len(self.Artist.filter(id='id')) == 0

    def test_by_ids_some_objects_valid_filter(self):
        a = self.Artist.create()
        self.Artist.create()
        objs = self.Artist.filter([a['id']])
        assert len(objs) == 1
        assert isinstance(objs[0], self.Artist)
        assert objs[0]['id'] == a['id']

    def test_by_kwargs_some_objects_valid_filter(self):
        a = self.Artist.create()
        self.Artist.create()
        objs = self.Artist.filter(id=a['id'])
        assert len(objs) == 1
        assert isinstance(objs[0], self.Artist)
        assert objs[0]['id'] == a['id']

    def test_by_ids_some_objects_deleted_valid_filter(self):
        a = self.Artist.create()
        a_id = a['id']
        self.Artist.create()
        a.delete()
        assert len(self.Artist.filter([a_id])) == 0

    def test_by_kwargs_some_objects_deleted_valid_filter(self):
        a = self.Artist.create()
        a_id = a['id']
        self.Artist.create()
        a.delete()
        assert len(self.Artist.filter(id=a_id)) == 0

    def test_by_ids_some_objects_invalid_filter(self):
        self.Artist.create()
        self.Artist.create()
        assert len(self.Artist.filter(['id'])) == 0

    def test_by_kwargs_some_objects_invalid_filter(self):
        self.Artist.create()
        self.Artist.create()
        assert len(self.Artist.filter(id='id')) == 0


class CountTests(DbBaseTestCase):
    def setUp(self):
        super(CountTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

        create_tables()
        create_indexes()

    def test_no_objects(self):
        assert self.Artist.count() == 0

    def test_some_objects(self):
        self.Artist.create()
        self.Artist.create()
        assert self.Artist.count() == 2


class WrapTests(DbBaseTestCase):
    def setUp(self):
        super(WrapTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

        create_tables()
        create_indexes()

    def test_correct_model_cls(self):
        obj = self.Artist.objects._wrap({})
        assert isinstance(obj, self.Artist)

    def test_correct_fields(self):
        doc = {'name': 'Andrei'}
        obj = self.Artist.objects._wrap(doc)
        assert obj.fields.__dict__ == doc


class ObjectSetTests(DbBaseTestCase):
    def setUp(self):
        super(ObjectSetTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

        create_tables()
        create_indexes()

    def test_cache_set_by_call(self):
        objs = self.Artist.all()
        # Cache should be None; no queries have been run
        assert objs.result_cache is None
        # Run query
        len(objs)
        # Cache should have been populated by run query
        assert objs.result_cache is not None

    def test_same_cache_between_calls(self):
        objs = self.Artist.all()
        len(objs)
        result_cache = objs.result_cache
        # Another call should hit the cache instead of running the query again
        len(objs)
        assert objs.result_cache == result_cache


class LenTests(DbBaseTestCase):
    def setUp(self):
        super(LenTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

        create_tables()
        create_indexes()

    def test_no_objects(self):
        assert len(self.Artist.all()) == 0

    def test_some_objects(self):
        self.Artist.create()
        self.Artist.create()
        assert len(self.Artist.all()) == 2

    def test_some_objects_deleted(self):
        a = self.Artist.create()
        self.Artist.create()
        a.delete()
        assert len(self.Artist.all()) == 1


class GetItemTests(DbBaseTestCase):
    def setUp(self):
        super(GetItemTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

        create_tables()
        create_indexes()

    def test_no_objects(self):
        with pytest.raises(IndexError):
            self.Artist.all()[0]

    def test_one_object(self):
        a = self.Artist.create()
        assert self.Artist.all()[0]['id'] == a['id']

    def test_some_objects(self):
        a1 = self.Artist.create()
        a2 = self.Artist.create()
        assert self.Artist.all()[0]['id'] in (a1['id'], a2['id'])
        assert self.Artist.all()[1]['id'] in (a1['id'], a2['id'])

    def test_invalid_index(self):
        self.Artist.create()
        with pytest.raises(IndexError):
            self.Artist.all()[1]

    # TODO: Write tests for negative indices, slices


class CustomQueryTests(DbBaseTestCase):
    def setUp(self):
        super(CustomQueryTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

        create_tables()
        create_indexes()

    def test_query_correctly_handled(self):
        self.Artist.create(name='John')
        self.Artist.create(name='Andrei')
        with get_conn() as conn:
            # order_by should be correctly handled by self.Artist
            results = list(self.Artist.order_by('name').run(conn))
        assert results[0]['name'] == 'Andrei'
        assert results[1]['name'] == 'John'



def callback(name):
    def _decorator(func):
        setattr(func, name, True)
        return func
    return _decorator


class cached_property(object):
    def __init__(self, func):
        self.func = func
        self.name = func.__name__

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        # Set an instance attribute having the same name as the cached property
        # so that the descriptor won't be called again when accessing the property
        # (see how descriptors work)
        result = instance.__dict__[self.name] = self.func(instance)
        return result


class classproperty(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner=None):
        return self.func(owner)


class classaccessonlyproperty(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner=None):
        if instance is not None:
            raise AttributeError('Property "%s" is not available via "%s" instances'
                                 % (self.func.__name__, owner.__name__))
        return self.func(owner)


class classaccessonly(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner=None):
        if instance is not None:
            raise AttributeError('Method is not available via %s instances'
                                 % owner.__name__)
        def newfunc(*args, **kwargs):
            return self.func(owner, *args, **kwargs)
        return newfunc


def synchronized(lock):
    def wrap(func):
        def synchronized_func(*args, **kwargs):
            with lock:
                return func(*args, **kwargs)
        return synchronized_func
    return wrap

class dispatch_to_metaclass(object):
    """
    Dispatches calls on a model method (which happens to be named the same as an
    instance method) to its metaclass. This has to be done since Python sticks
    to an instance method even though a call was made to a class method if they
    both share the same name.
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, instance, model=None):
        if instance is None:
            metaclass = type(model)
            return getattr(metaclass, self.func.__name__).__get__(model, metaclass)
        return self.func.__get__(instance, model)



import rethinkdb as r

import remodel.connection


run = r.ast.RqlQuery.run

def remodel_run(self, c=None, **global_optargs):
    """
    Passes a connection from the connection pool so that we can call .run()
    on a query without an explicit connection
    """

    if not c:
        with remodel.connection.get_conn() as conn:
            return run(self, conn, **global_optargs)
    else:
        return run(self, c, **global_optargs)

r.ast.RqlQuery.run = remodel_run



import rethinkdb as r


def create_tables():
    from .registry import model_registry

    created_tables = r.table_list().run()
    for model_cls in model_registry.all().values():
        if model_cls._table not in created_tables:
            result = r.table_create(model_cls._table).run()
            if result['tables_created'] != 1:
                raise RuntimeError('Could not create table %s for model %s' % (
                                   model_cls._table, model_cls.__name__))


def drop_tables():
    from .registry import model_registry

    created_tables = r.table_list().run()
    for model_cls in model_registry.all().values():
        if model_cls._table in created_tables:
            result = r.table_drop(model_cls._table).run()
            if result['tables_dropped'] != 1:
                raise RuntimeError('Could not drop table %s for model %s' % (
                                   model_cls._table, model_cls.__name__))


def create_indexes():
    from .registry import model_registry, index_registry

    for model, index_set in index_registry.all().items():
        model_cls = model_registry.get(model)
        created_indexes = r.table(model_cls._table).index_list().run()
        for index in index_set:
            if index not in created_indexes:
                result = r.table(model_cls._table).index_create(index).run()
                if result['created'] != 1:
                    raise RuntimeError('Could not create index %s for table %s' % (
                                       index, model_cls._table))



import rethinkdb as r
from six import add_metaclass
from inflection import tableize

from .decorators import callback, classaccessonlyproperty, dispatch_to_metaclass
from .errors import OperationError
from .field_handler import FieldHandlerBase, FieldHandler
from .object_handler import ObjectHandler
from .registry import model_registry
from .utils import deprecation_warning


REL_TYPES = ('has_one', 'has_many', 'belongs_to', 'has_and_belongs_to_many')
CALLBACKS = ('before_save', 'after_save', 'before_delete', 'after_delete', 'after_init')


class ModelBase(type):
    def __new__(mcs, name, bases, dct):
        super_new = super(ModelBase, mcs).__new__

        # Ensure the following are not done for the Model class itself
        parents = [b for b in bases if isinstance(b, ModelBase)]
        if not parents:
            return super_new(mcs, name, bases, dct)

        # Set metadata
        dct['_table'] = tableize(name)

        rel_attrs = {rel: dct.setdefault(rel, ()) for rel in REL_TYPES}
        dct['_field_handler_cls'] = FieldHandlerBase(
            '%sFieldHandler' % name,
            (FieldHandler,),
            dict(rel_attrs, model=name))
        object_handler_cls = dct.setdefault('object_handler', ObjectHandler)

        # Register callbacks
        dct['_callbacks'] = {callback: [] for callback in CALLBACKS}
        for callback in CALLBACKS:
            # Callback-named methods
            if callback in dct:
                dct['_callbacks'][callback].append(callback)
            # Callback-decorated methods
            dct['_callbacks'][callback].extend([key
                                                for key, value in dct.items()
                                                if hasattr(value, callback)])

        new_class = super_new(mcs, name, bases, dct)
        model_registry.register(name, new_class)
        setattr(new_class, 'objects', object_handler_cls(new_class))
        return new_class

    # Proxies undefined attributes to Model.objects; useful for building
    # ReQL queries directly on the Model (e.g.: User.order_by('name').run())
    def __getattr__(self, name):
        return getattr(self.objects, name)


@add_metaclass(ModelBase)
class Model(object):
    def __init__(self, **kwargs):
        self.fields = self._field_handler_cls()

        for key, value in kwargs.items():
            # Assign fields this way to be sure that validation takes place
            setattr(self.fields, key, value)

        self._run_callbacks('after_init')

    def save(self):
        self._run_callbacks('before_save')

        fields_dict = self.fields.as_dict()
        try:
            # Attempt update
            id_ = fields_dict['id']
            result = (r.table(self._table).get(id_).replace(r.row
                        .without(r.row.keys().difference(list(fields_dict.keys())))
                        .merge(fields_dict), return_changes=True).run())

        except KeyError:
            # Resort to insert
            result = (r.table(self._table).insert(fields_dict, return_changes=True)
                      .run())

        if result['errors'] > 0:
            raise OperationError(result['first_error'])

        # RethinkDB 2.0 doesn't add the 'changes' key in the result if the
        # document hasn't been modified
        # TODO: Follow on the discussion at linkyndy/remodel#23 and change this
        #       accordingly
        if 'changes' in result:
            # Force overwrite so that related caches are flushed
            self.fields.__dict__ = result['changes'][0]['new_val']

        self._run_callbacks('after_save')

    def update(self, **kwargs):
        for key, value in kwargs.items():
            # Assign fields this way to be sure that validation takes place
            setattr(self.fields, key, value)

        self.save()

    def delete(self):
        self._run_callbacks('before_delete')

        try:
            id_ = getattr(self.fields, 'id')
            result = r.table(self._table).get(id_).delete().run()
        except AttributeError:
            raise OperationError('Cannot delete %r (object not saved or '
                                 'already deleted)' % self)

        if result['errors'] > 0:
            raise OperationError(result['first_error'])

        delattr(self.fields, 'id')
        # Remove any reference to the deleted object
        for field in self.fields.related:
            delattr(self.fields, field)

        self._run_callbacks('after_delete')

    # TODO: Get rid of this nasty decorator after renaming .get() on ObjectHandler
    @dispatch_to_metaclass
    def get(self, key, default=None):
        try:
            return getattr(self.fields, key)
        except AttributeError:
            return default

    def __getitem__(self, key):
        try:
            return getattr(self.fields, key)
        except AttributeError:
            raise KeyError(key)

    def __setitem__(self, key, value):
        try:
            setattr(self.fields, key, value)
        except AttributeError:
            raise KeyError(key)

    def __delitem__(self, key):
        try:
            delattr(self.fields, key)
        except AttributeError:
            raise KeyError(key)

    def __contains__(self, item):
        return hasattr(self.fields, item)

    def __repr__(self):
        try:
            id_ = self.fields.id
        except AttributeError:
            id_ = 'not saved'
        return '<%s: %s>' % (self.__class__.__name__, id_)

    def __str__(self):
        return '<%s object>' % self.__class__.__name__

    def _run_callbacks(self, name):
        for callback in self._callbacks[name]:
            getattr(self, callback)()

    @classaccessonlyproperty
    def table(self):
        deprecation_warning('Model.table will be deprecated soon. Please use '
                            'Model.objects to build any custom query on a '
                            'Model\'s table (read more about ObjectHandler)')
        return self.objects


before_save = callback('before_save')
after_save = callback('after_save')
before_delete = callback('before_delete')
after_delete = callback('after_delete')
after_init = callback('after_init')



import rethinkdb as r


class ObjectHandler(object):
    def __init__(self, model_cls, query=None):
        self.model_cls = model_cls
        self.query = query or r.table(model_cls._table)

    def __getattr__(self, name):
        return getattr(self.query, name)

    def all(self):
        return ObjectSet(self, self.query)

    def create(self, **kwargs):
        obj = self.model_cls(**kwargs)
        obj.save()
        return obj

    def get(self, id_=None, **kwargs):
        if id_:
            try:
                doc = self.query.get(id_).run()
            except AttributeError:
                # self.query has a get_all applied, cannot call get
                kwargs.update(id=id_)
            else:
                if doc is not None:
                    return self._wrap(doc)
                return None
        docs = self.query.filter(kwargs).limit(1).run()
        try:
            return self._wrap(list(docs)[0])
        except IndexError:
            return None

    def get_or_create(self, id_=None, **kwargs):
        obj = self.get(id_, **kwargs)
        if obj:
            return obj, False
        return self.create(**kwargs), True

    def filter(self, ids=None, **kwargs):
        if ids:
            try:
                query = self.query.get_all(r.args(ids)).filter(kwargs)
            except AttributeError:
                # self.query already has a get_all applied
                query = (self.query.filter(lambda doc: r.expr(ids).contains(doc['id']))
                                   .filter(kwargs))
        else:
            query = self.query.filter(kwargs)
        return ObjectSet(self, query)

    def count(self):
        return self.query.count().run()

    def _wrap(self, doc):
        obj = self.model_cls()
        # Assign fields this way to skip validation
        obj.fields.__dict__ = doc
        return obj


class ObjectSet(object):
    def __init__(self, object_handler, query):
        self.object_handler = object_handler
        self.query = query
        self.result_cache = None

    def __iter__(self):
        self._fetch_results()
        return iter(self.result_cache)

    def __len__(self):
        self._fetch_results()
        return len(self.result_cache)

    def __getitem__(self, key):
        self._fetch_results()
        return self.result_cache[key]

    def iterator(self):
        results = self.query.run()
        for doc in results:
            yield self.object_handler._wrap(doc)

    def _fetch_results(self):
        if self.result_cache is None:
            self.result_cache = list(self.iterator())



"""
remodel
----------------

Very simple yet powerful and extensible Object Document Mapper for RethinkDB, written in Python.

"""

from setuptools import setup, find_packages


setup(
    name='remodel',
    version='0.4.1',
    url='http://github.com/linkyndy/remodel',
    license='MIT',
    author='Andrei Horak',
    author_email='linkyndy@gmail.com',
    description='RethinkDB ODM',
    long_description=__doc__,
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'rethinkdb',
        'inflection',
        'six'
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



from collections import defaultdict

from .errors import AlreadyRegisteredError
import remodel.models


class ModelRegistry(object):
    def __init__(self):
        self._data = {}

    def __len__(self):
        return len(self._data)

    def register(self, name, cls):
        if name in self._data:
            raise AlreadyRegisteredError('Model "%s" has been already registered' % name)
        if not issubclass(cls, remodel.models.Model):
            raise ValueError('Registered model class "%r" must be a subclass of "Model"' % cls)
        self._data[name] = cls

    def unregister(self, name):
        if name not in self._data:
            raise KeyError('"%s" is not a registered model' % name)
        del self._data[name]

    def get(self, name):
        if name not in self._data:
            raise KeyError('Model "%s" has not been registered' % name)
        return self._data[name]

    def all(self):
        return self._data

    def clear(self):
        self._data = {}


# Used by ModelBase metaclass to register every declared Model class.
model_registry = ModelRegistry()


class IndexRegistry(object):
    def __init__(self):
        self._data = defaultdict(set)

    def register(self, model, index):
        self._data[model].add(index)

    def unregister(self, model, index):
        self._data[model].discard(index)

    def get_for_model(self, model):
        if model not in self._data:
            return set()
        return self._data[model]

    def all(self):
        return self._data

    def clear(self):
        self._data = defaultdict(set)


index_registry = IndexRegistry()



import pytest
import rethinkdb as r

from remodel.errors import OperationError
from remodel.helpers import create_tables, create_indexes
from remodel.models import Model, before_save, after_save, before_delete, after_delete, after_init
from remodel.object_handler import ObjectHandler
from remodel.registry import model_registry
from remodel.related import (HasOneDescriptor, BelongsToDescriptor,
                             HasManyDescriptor, HasAndBelongsToManyDescriptor)

from . import BaseTestCase, DbBaseTestCase


class ModelTests(BaseTestCase):
    def test_model_attributes(self):
        class Artist(Model):
            pass

        assert hasattr(Artist, '_table')
        assert hasattr(Artist, 'has_one')
        assert hasattr(Artist, 'has_many')
        assert hasattr(Artist, 'has_and_belongs_to_many')
        assert hasattr(Artist, 'belongs_to')
        assert hasattr(Artist, '_field_handler_cls')
        assert hasattr(Artist, 'object_handler')
        assert hasattr(Artist, '_callbacks')
        assert hasattr(Artist, 'objects')

    def test_default_object_handler_cls(self):
        class Artist(Model):
            pass

        assert Artist.object_handler == ObjectHandler
        assert isinstance(Artist.objects, ObjectHandler)

    def test_custom_object_handler_cls(self):
        class CustomObjectHandler(ObjectHandler):
            pass

        class Artist(Model):
            object_handler = CustomObjectHandler

        assert Artist.object_handler == CustomObjectHandler
        assert isinstance(Artist.objects, CustomObjectHandler)

    def test_callback_keys(self):
        """
        Test whether all callback keys are created for model callback registry
        """

        class Artist(Model):
            pass

        assert isinstance(Artist._callbacks, dict)
        assert 'before_save' in Artist._callbacks
        assert 'after_save' in Artist._callbacks
        assert 'before_delete' in Artist._callbacks
        assert 'after_delete' in Artist._callbacks
        assert 'after_init' in Artist._callbacks

    def test_callback_named_methods_registered(self):
        """
        Test whether all callback named methods are registered to the model
        callback registry
        """

        class Artist(Model):
            def before_save(self):
                pass

            def after_save(self):
                pass

            def before_delete(self):
                pass

            def after_delete(self):
                pass

            def after_init(self):
                pass

        assert 'before_save' in Artist._callbacks['before_save']
        assert 'after_save' in Artist._callbacks['after_save']
        assert 'before_delete' in Artist._callbacks['before_delete']
        assert 'after_delete' in Artist._callbacks['after_delete']
        assert 'after_init' in Artist._callbacks['after_init']

    def test_callback_decorated_methods_registered(self):
        """
        Test whether all callback decorated methods are registered to the model
        callback registry
        """

        class Artist(Model):
            @before_save
            def bef_save(self):
                pass

            @after_save
            def aft_save(self):
                pass

            @before_delete
            def bef_delete(self):
                pass

            @after_delete
            def aft_delete(self):
                pass

            @after_init
            def aft_init(self):
                pass

        assert 'bef_save' in Artist._callbacks['before_save']
        assert 'aft_save' in Artist._callbacks['after_save']
        assert 'bef_delete' in Artist._callbacks['before_delete']
        assert 'aft_delete' in Artist._callbacks['after_delete']
        assert 'aft_init' in Artist._callbacks['after_init']


class FieldTests(BaseTestCase):
    """
    Tests both whether fields passed as keyword arguments on a new document
    and fields passed after document creation are correctly set and retrieved
    """

    def setUp(self):
        super(FieldTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

    def test_new_document_with_no_fields(self):
        a = self.Artist()

    def test_new_document_with_one_field(self):
        a = self.Artist(name='Andrei')
        assert 'name' in a
        assert a['name'] == 'Andrei'

    def test_new_document_with_several_fields(self):
        a = self.Artist(name='Andrei', country='Romania', height='tall')
        assert 'name' in a and 'country' in a and 'height' in a
        assert (a['name'] == 'Andrei' and
                a['country'] == 'Romania' and
                a['height'] == 'tall')

    def test_existing_document(self):
        a = self.Artist()
        a['name'] = 'Andrei'
        assert 'name' in a
        assert a['name'] == 'Andrei'


class RelatedFieldTests(DbBaseTestCase):
    """
    Tests both whether related fields passed as keyword arguments on a new document
    and related fields passed after document creation are correctly set and retrieved
    """

    def setUp(self):
        super(RelatedFieldTests, self).setUp()

        class Artist(Model):
            has_one = ('Bio',)
            belongs_to = ('Person',)
            has_many = ('Song',)
            has_and_belongs_to_many = ('Taste',)
        self.Artist = Artist

        class Bio(Model):
            pass
        self.Bio = Bio

        class Person(Model):
            pass
        self.Person = Person

        class Song(Model):
            pass
        self.Song = Song

        class Taste(Model):
            pass
        self.Taste = Taste

        create_tables()
        create_indexes()

    def test_existing_document_with_has_one_field(self):
        a = self.Artist()
        a.save()
        b = self.Bio()
        a['bio'] = b
        assert a['bio'] is b

    def test_new_document_with_belongs_to_field(self):
        p = self.Person()
        p.save()
        a = self.Artist(person=p)
        assert a['person'] is p

    def test_existing_document_with_belongs_to_field(self):
        p = self.Person()
        p.save()
        a = self.Artist()
        a['person'] = p
        assert a['person'] is p

    def test_existing_document_with_has_many_field(self):
        a = self.Artist()
        a.save()
        s1 = self.Song()
        s2 = self.Song()
        a['songs'] = [s1, s2]

    def test_existing_document_with_has_and_belongs_to_many_field(self):
        a = self.Artist()
        a.save()
        t1 = self.Taste()
        t1.save()
        t2 = self.Taste()
        t2.save()
        a['tastes'] = [t1, t2]


class SaveTests(DbBaseTestCase):
    def setUp(self):
        super(SaveTests, self).setUp()

        class Artist(Model):
            has_one = ('Bio',)
            belongs_to = ('Person',)
        self.Artist = Artist

        class Bio(Model):
            pass
        self.Bio = Bio

        class Person(Model):
            pass
        self.Person = Person

        create_tables()
        create_indexes()

    def assert_saved(self, table, fields):
        assert len(list(r.table(table).filter(fields).run())) == 1

    def test_insert(self):
        a = self.Artist(name='Andrei')
        a.save()
        self.assert_saved(a._table, a.fields.as_dict())

    def test_update(self):
        a = self.Artist(name='Andrei')
        a.save()
        a.save()
        self.assert_saved(a._table, a.fields.as_dict())

    def test_update_with_added_field(self):
        a = self.Artist(name='Andrei')
        a.save()
        a['country'] = 'Romania'
        a.save()
        self.assert_saved(a._table, a.fields.as_dict())

    def test_update_with_removed_field(self):
        a = self.Artist(name='Andrei')
        a.save()
        del a['name']
        a.save()
        self.assert_saved(a._table, a.fields.as_dict())

    def test_belongs_to(self):
        p = self.Person()
        p.save()
        a = self.Artist()
        a['person'] = p
        a.save()
        self.assert_saved(a._table, a.fields.as_dict())

    def test_updated_belongs_to(self):
        p1 = self.Person()
        p1.save()
        a = self.Artist()
        a['person'] = p1
        a.save()
        p2 = self.Person()
        p2.save()
        a['person'] = p2
        a.save()
        self.assert_saved(a._table, a.fields.as_dict())

    def test_removed_belongs_to(self):
        p = self.Person()
        p.save()
        a = self.Artist()
        a['person'] = p
        a.save()
        del a['person']
        a.save()
        self.assert_saved(a._table, a.fields.as_dict())

    def test_has_one(self):
        a = self.Artist()
        a.save()
        b = self.Bio()
        a['bio'] = b
        b.save()
        self.assert_saved(b._table, b.fields.as_dict())

    def test_updated_has_one(self):
        a = self.Artist()
        a.save()
        b1 = self.Bio()
        a['bio'] = b1
        b1.save()
        b2 = self.Bio()
        a['bio'] = b2
        b1.save()
        self.assert_saved(b1._table, b1.fields.as_dict())

    def test_removed_has_one(self):
        a = self.Artist()
        a.save()
        b = self.Bio()
        a['bio'] = b
        b.save()
        del a['bio']
        b.save()
        self.assert_saved(b._table, b.fields.as_dict())


class DeleteTests(DbBaseTestCase):
    def setUp(self):
        super(DeleteTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

        create_tables()
        create_indexes()

    def assert_deleted(self, table, fields):
        assert len(list(r.table(table).filter(fields).run())) == 0

    def test_not_saved(self):
        a = self.Artist()
        with pytest.raises(OperationError):
            a.delete()

    def test_already_deleted(self):
        a = self.Artist()
        a.save()
        a.delete()
        with pytest.raises(OperationError):
            a.delete()

    # TODO: Add tests for confirming that related objects have no reference left to the deleted object


class GetTests(BaseTestCase):
    def setUp(self):
        super(GetTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

    def test_existing_field(self):
        a = self.Artist(name='Andrei')
        assert a.get('name') == 'Andrei'

    def test_inexisting_field_with_default(self):
        a = self.Artist()
        assert a.get('name', 'Andrei') == 'Andrei'

    def test_inexisting_field_without_default(self):
        a = self.Artist()
        assert a.get('name') is None


class UpdateTests(DbBaseTestCase):
    def setUp(self):
        super(UpdateTests, self).setUp()

        class Artist(Model):
            pass
        self.Artist = Artist

        create_tables()
        create_indexes()

    def assert_updated(self, table, fields):
        assert len(list(r.table(table).filter(fields).run())) == 1

    def test_one_field_updated(self):
        a = self.Artist()
        a.update(name='Andrei')
        assert 'name' in a
        assert a['name'] == 'Andrei'

    def test_several_fields_updated(self):
        a = self.Artist()
        a.update(name='Andrei', country='Romania', height='tall')
        assert 'name' in a and 'country' in a and 'height' in a
        assert (a['name'] == 'Andrei' and
                a['country'] == 'Romania' and
                a['height'] == 'tall')

    def test_new_document_with_no_fields(self):
        a = self.Artist(name='Andrei')
        a.update()
        self.assert_updated(a._table, a.fields.as_dict())

    def test_new_document_with_one_field(self):
        a = self.Artist(name='Andrei')
        a.update(country='Romania')
        self.assert_updated(a._table, a.fields.as_dict())

    def test_new_document_with_several_fields(self):
        a = self.Artist(name='Andrei')
        a.update(country='Romania', city='Timisoara', male=True)
        self.assert_updated(a._table, a.fields.as_dict())

    def test_existing_document_with_no_fields(self):
        a = self.Artist.create(name='Andrei')
        a.update()
        self.assert_updated(a._table, a.fields.as_dict())

    def test_existing_document_with_one_field(self):
        a = self.Artist.create(name='Andrei')
        a.update(country='Romania')
        self.assert_updated(a._table, a.fields.as_dict())

    def test_existing_document_with_several_fields(self):
        a = self.Artist.create(name='Andrei')
        a.update(country='Romania', city='Timisoara', male=True)
        self.assert_updated(a._table, a.fields.as_dict())


class CallbackTests(DbBaseTestCase):
    """
    Tests whether callbacks are run and also that they are run at the desired
    moment
    """

    # TODO: Make these tests more relevant by using mocks

    def assert_saved(self, table, fields):
        assert len(list(r.table(table).filter(fields).run())) == 1

    def assert_not_saved(self, table, fields):
        assert len(list(r.table(table).filter(fields).run())) == 0

    def assert_deleted(self, table, fields):
        assert len(list(r.table(table).filter(fields).run())) == 0

    def test_before_save(self):
        class Artist(Model):
            def before_save(self):
                self['verified'] = True

        create_tables()

        a = Artist()
        a.save()
        assert a['verified'] is True
        self.assert_saved(a._table, a.fields.as_dict())

    def test_after_save(self):
        class Artist(Model):
            def after_save(self):
                self['confirmed'] = False

        create_tables()

        a = Artist()
        a.save()
        assert a['confirmed'] is False
        self.assert_not_saved(a._table, a.fields.as_dict())

    def test_before_delete(self):
        class Artist(Model):
            def before_delete(self):
                self['deleted'] = True

        create_tables()

        a = Artist.create()
        a.delete()
        assert a['deleted'] is True
        self.assert_deleted(a._table, a.fields.as_dict())

    def test_after_delete(self):
        class Artist(Model):
            def after_delete(self):
                self['deleted'] = True

        create_tables()

        a = Artist.create()
        a.delete()
        assert a['deleted'] is True
        self.assert_deleted(a._table, a.fields.as_dict())

    def test_after_init(self):
        class Artist(Model):
            def after_init(self):
                self['new'] = True

        a = Artist()
        assert a['new'] is True
